unit CRPrintX;

interface

uses
  SysUtils, WinTypes, WinProcs, DbiErrs, DbiTypes, DbiProcs, Printers,
  Classes, Controls, Graphics, Mask, DB, Forms, Dialogs, Messages;

const
(* Supporting Objects *)
  (* GENERAL          *)
  PE_WORD_LEN              = 2;
  PE_DWORD_LEN             = 4;
  PE_DLL_NAME_LEN          = 64;
  PE_FULL_NAME_LEN         = 256;
  (* TABLE TYPE       *)
  PE_SIZEOF_TABLE_TYPE     = PE_WORD_LEN + PE_DLL_NAME_LEN + PE_FULL_NAME_LEN + PE_WORD_LEN;
  (* TABLE LOCATION   *)
  PE_TABLE_LOCATION_LEN    = 256;
  PE_SIZEOF_TABLE_LOCATION = PE_WORD_LEN + PE_TABLE_LOCATION_LEN;
  (* WINDOW STATE INFORMATION *)
  WSE_MINIMIZE             = 536870912;
  WSE_VISIBLE              = 268435456;
  WSE_MAXIMIZE             = 16777216;
  WSE_CAPTION              = 12582912;
  WSE_SYSMENU              = 524288;
  WSE_THICKFRAME           = 262144;
  WSE_MINBOX               = 131072;
  WSE_MAXBOX               = 65536;
  WSE_BASIS                = WSE_VISIBLE + WSE_CAPTION + WSE_SYSMENU + WSE_THICKFRAME + WSE_MAXBOX + WSE_MINBOX;
  (* WINDOW SIZE & PLACEMENT INFORMATION *)
  WSE_USEDEFAULT           = -32768;

type
(* ************************************************************************** *)
(* Supporting Objects *)
  TXDestType = (toFile, toPrinter, toWindow);
  TXPrinterStatus = (psOk, psUnassociated, psInvalidDevice);
  TXWinLocal = (wpDefault, wpDesign);
  TXWinStyle = (wslNormal, wslMDIChild);
  TXWinState = (wstNormal, wstMinimized, wstMaximized);

(* ************************************************************************** *)
(* VCL Object *)
  TCrystalVCL = class(TComponent)
    private      (* Private declarations   *)
      (* Windows Printer Zone Specific Data                                   *)
      FDevice             : PChar;           (* Printer Description           *)
      FDriver             : PChar;           (* Printer Description           *)
      FPort               : PChar;           (* Printer Description           *)
      FStatus             : TXPrinterStatus; (* Device Status                 *)
                                             (*  Error : Status <> psOk       *)
      FDeviceModule       : THandle;         (* Printer Module Handle         *)
      FDeviceMode         : TDeviceMode;     (* FarProc : DevMode             *)
      FExtDeviceMode      : TExtDeviceMode;  (* FarProc : ExtDevMode          *)
      FDevSettings        : ^TDevMode;       (* Printer Modes - Copy          *)
      FDevSettingSize     : Integer;         (* Size of Printer Modes         *)
      FAUPrinterSpec      : Boolean;         (* Auto Update Printer Spec.     *)
      (* Crystal Report VCL Specific Data                                     *)
      FOnLine             : Boolean;         (* General Status Control        *)
      FBDEerror           : Boolean;         (* Borland Database Engine       *)
      FCREerror           : Boolean;         (* Crystal Reports Engine        *)
      (* Crystal Report Engine Specific Data                                  *)
      FReportFile         : string[255];     (* Name of Report File           *)
      FReportTitle        : string;          (* Major Report Title            *)
      FMinorTitle         : string;          (* Minor Report Title            *)
      FWindowCaption      : string[100];     (* Report Window Caption         *)
      FWindowLocal        : TXWinLocal;      (* Window Position Control       *)
      FWindowStyle        : TXWinStyle;      (* Style: wslMDIChild /wslNormal *)
      FWindowState        : TXWinState;      (* State: wstNorm /wstMin /wstMax*)
      FWindowArea         : TRect;           (* Actual Window Dimensions      *)
      FWindow             : TRect;           (* Window Dimemsion Interface    *)
      FLocation           : TStrings;        (* Table Locations               *)
      FLocationOutOfDate  : Boolean;         (* Flags OOD Status              *)
      FSelFormula         : TStrings;        (* Selection Formula             *)
      FErrorCode          : LongInt;         (* VCL Error Code Holder         *)
                                             (*  AKA.  EXECUTE Result         *)
      FErrorMessage         : string;        (* VCL Error Text Holder         *)
    protected    (* Protected declarations *)
      function GetReportFile : string;
      function GetReportTitle(Index: Integer) : string;
      function GetWindowTitle : string;
      function GetWindowItem(Index: Integer) : Longint;
      procedure SetReportFile(DReportFile: string);
      procedure SetReportTitle(Index: Integer; DTitle: string);
      procedure SetWindowTitle(DTitle: string);
      procedure SetWindowLocal(DValue: TXWinLocal);
      procedure SetWindowStyle(DValue: TXWinStyle);
      procedure SetWindowItem(Index: Integer; DValue: Longint);
      procedure SetString(Index: Integer; Value: TStrings);
    private
      (* Windows Printer Zone Specific Routines                          *)
      function DecodeSTR(var Str: PChar): PChar;
      procedure ClearPrinterDevice;
      procedure GetPrinterDeviceSpec(ADevice, ADriver, APort: PChar);
      (* Crystal Report Engine Specific Data                             *)
      (* Update Report Base Table Location(s) to specified Aliases       *)
      procedure DetermineError(var printJob: Integer);
      function UpdateReportLocation(var printJob: Integer) : Boolean;
      function GetAliasLocation(const AliasName: string) : string;
      procedure GetAddressComponents(DAddress: string; var DName: string;
                                     var DPath: string);
      function GetNewLocation(FilePath: string) : string;
      (* Update General Formulas with User-Definition                    *)
      function UpdateSmallFormula(var printJob: Integer;
                                  DFName, DFormula: string) : Boolean;
      (* Update Record Selection Criteria with User-Definition        *)
      function UpdateSelectionFormula(var printJob: Integer) : Boolean;
      function GetStringSize(DStringList: TStrings) : LongInt;
    public       (* Public declarations    *)
      constructor Create(AOwner: TComponent); override;
      destructor Destroy; override;
      (* Crystal Report Engine Specific Data                             *)
      function Execute : Boolean;
      function ConvertAliases : Boolean;
      (* Windows Printer Zone Specific Routines                          *)
      function UpdatePrinterDeviceSpec : Boolean;
    published    (* Published declarations *)
      property ActivePrinter  : Boolean    read FAUPrinterSpec write FAUPrinterSpec;
      property ErrorCode      : LongInt    read FErrorCode     write FErrorCode;
      property ErrorMessage   : string     read FErrorMessage  write FErrorMessage;
      property ReportFile     : string     read GetReportFile  write SetReportFile;
      property ReportTitle    : string     index  0   read GetReportTitle write SetReportTitle;
      property MinorTitle     : string     index  1   read GetReportTitle write SetReportTitle;
      property WindowCaption  : string     read GetWindowTitle write SetWindowTitle;
      property WindowPosition : TXWinLocal read FWindowLocal   write SetWindowLocal;
      property WindowStyle    : TXWinStyle read FWindowStyle   write SetWindowStyle;
      property WindowState    : TXWinState read FWindowState   write FWindowState;
      property WindowTop      : LongInt    index  0   read GetWindowItem  write SetWindowItem;
      property WindowLeft     : LongInt    index  1   read GetWindowItem  write SetWindowItem;
      property WindowWidth    : LongInt    index  2   read GetWindowItem  write SetWindowItem;
      property WindowHeight   : LongInt    index  3   read GetWindowItem  write SetWindowItem;
      property Location       : TStrings   index  0   read FLocation      write SetString;
      property SelectFormula  : TStrings   index  1   read FSelFormula    write SetString;
   end;


(* ************************************************************************** *)
(* ************************************************************************** *)

(* Supporting Data-Types and Structures *)
  PEDllNameType       = array [1..PE_DLL_NAME_LEN] of Char;
  PEFullNameType      = array [1..PE_FULL_NAME_LEN] of Char;
  PETableLocType      = array [1..PE_TABLE_LOCATION_LEN] of Char;

  PEJobInfo = record
                StructSize        : Word;    (*initialize to sizeof (PEJobInfo)*)
                NumRecordsRead    : LongInt;
                NumRecordsSelected: LongInt;
                NumRecordsPrinted : LongInt;
                DisplayPageN      : Word;    (*the page being displayed in window*)
                LatestPageN       : Word;    (*the page being generated*)
                StartPageN        : Word;    (*user opted, default to 1*)
                PrintEnded        : Boolean; (*full report print completed?*)
               end;

  PETableType = record
                  StructSize      : Word;           (* INIT: PE_SIZEOF_TABLE_TYPE  *)
                  DLLName         : PEDllNameType;  (* Strings are null-terminated *)
                  DescriptiveName : PEFullNameType;
                  DBType          : Word;
                 end;

  PETableLocation = record
                      StructSize  : Word;           (* INIT: PE_SIZEOF_TABLE_LOCATION  *)
                      Location    : PETableLocType; (* String is null-terminated       *)
                     end;

(* ************************************************************************** *)

  (* Open and Close Print Engine *)
  function PEOpenEngine: Boolean;
  procedure PECloseEngine;
  (* Get Version Information *)
  function PEGetVersion(versionRequested: Integer) : Integer;

  (* Open, print & Close Report (used when no changes needed to report) *)
  function PEPrintReport(reportFilePath: PChar; toPrinter, toWindow: Boolean;
                         title: PChar; left, top, width, height: Integer;
                         style: LongInt; parentWindow: Hwnd) : Integer;

  (* Open and Close Print Job *)
  function PEOpenPrintJob(reportFilePath: PChar) : Integer;
  function PEClosePrintJob(printJob: Integer) : Boolean;

  (* Start and Cancel Print Job (print, usually after changing report) *)
  function PEStartPrintJob(printJob: Integer; waitUntilDone: Boolean) : Boolean;
  procedure PECancelPrintJob(printJob: Integer);

  (* Print Job Status *)
  function PEIsPrintJobFinished(printJob: Integer) : Boolean;
  function PEGetJobStatus(printJob: Integer; var jobInfo: PEJobInfo) : Integer;

  (* Print Job Error Codes and Messages routines *)
  function PEGetErrorCode (printJob: Integer) : Integer;
  function PEGetErrorText (printJob: Integer; var textHandle: HWnd;
                           var textLength: Integer) : Boolean;
  function PEGetHandleString (textHandle: HWnd; buffer: PChar;
                              bufferLength: Integer) : Boolean;

  (* Report Output Selection routine(s) *)
  function PEOutputToWindow (printJob: Integer; title: PChar;
                             left, top, width, height: Integer; style: LongInt;
                             parentWindow: HWnd) : Boolean;

  (* Specific Formula SET routine *)
  function PESetFormula (printJob: Integer; formulaName: PChar;
                         formulaString: PChar) : Boolean;

  (* Selection Criteria routines *)
  function PEGetSelectionFormula (printJob: Integer; var textHandle: HWnd;
                                  var textLength: Integer) : Boolean;
  function PESetSelectionFormula (printJob: Integer; formulaString: PChar):Boolean;
  function PECheckSelectionFormula (printJob: Integer) : Boolean;

  (* Database Determination & Location routines *)
  function PEGetNTables (printJob: Integer) : Integer;
  function PEGetNthTableType (printJob, tableN: Integer;
                              var tableType: PETableType) : Boolean;
  function PEGetNthTableLocation (printJob, tableN: Integer;
                                  var location: PETableLocation) : Boolean;
  function PESetNthTableLocation (printJob, tableN: Integer;
                                  var location: PETableLocation) : Boolean;

(* Changing printer selection *)
  function PESelectPrinter (printJob: Integer; driverName, printerName, portName: PChar;
                            mode: TDevMode) : Boolean;
  function PEGetSelectedPrinter (printJob: Integer;
                                 var driverHandle: Hwnd; var driverLength: Integer;
                                 var printerHandle: Hwnd; var printerLength: Integer;
                                 var portHandle: Hwnd; var portLength: Integer;
                                 var mode: TDevMode) : Boolean;


(* ************************************************************************** *)
(* ************************************************************************** *)

procedure Register;

implementation
{ $X+ }

(* ************************************************************************** *)
(* ************************************************************************** *)

procedure Register;
 begin
   RegisterComponents('Data Access', [TCrystalVCL]);
 end;

function  PEOpenEngine;                    external 'CRPE';
procedure PECloseEngine;                   external 'CRPE';
function  PEGetVersion;                    external 'CRPE';

function  PEPrintReport;                   external 'CRPE';

function  PEOpenPrintJob;                  external 'CRPE';
function  PEClosePrintJob;                 external 'CRPE';
function  PEStartPrintJob;                 external 'CRPE';
procedure PECancelPrintJob;                external 'CRPE';
function  PEIsPrintJobFinished;            external 'CRPE';
function  PEGetJobStatus;                  external 'CRPE';

function  PEGetErrorCode;                  external 'CRPE';
function  PEGetErrorText;                  external 'CRPE';
function  PEGetHandleString;               external 'CRPE';
{
function  PEOutputToPrinter;               external 'CRPE';
function  PEOutputToFile;                  external 'CRPE';
}
function  PEOutputToWindow;                external 'CRPE';

function  PESetFormula;                    external 'CRPE';

function  PEGetSelectionFormula;           external 'CRPE';
function  PESetSelectionFormula;           external 'CRPE';
function  PECheckSelectionFormula;         external 'CRPE';

function  PEGetNTables;                    external 'CRPE';
function  PEGetNthTableType;               external 'CRPE';
function  PEGetNthTableLocation;           external 'CRPE';
function  PESetNthTableLocation;           external 'CRPE';
{
function  PEGetWindowHandle;               external 'CRPE';
procedure PECloseWindow;                   external 'CRPE';
}
function PESelectPrinter;                  external 'CRPE';
function PEGetSelectedPrinter;             external 'CRPE';


(* ************************************************************************** *)
(* ************************************************************************** *)
(* ************************************************************************** *)
constructor TCrystalVCL.Create(AOwner: TComponent);
 begin
   inherited Create(AOwner);
   (* Windows Printer Zone Specific Routines                           *)
   FDevice := nil;
   FDriver := nil;
   FPort := nil;
   FDeviceModule := 0;
   FDevSettings := nil;
   FStatus := psUnassociated;
   GetPrinterDeviceSpec(nil, nil, nil);  (* Associate Default Printer  *)
   FAUPrinterSpec := True;  (* VCL resets Report Printer to Selection  *)
   (* Crystal Report Engine Specific Data                              *)
   FOnLine := Boolean(PEOpenEngine);
   FBDEerror := False;
   FCREerror := False;
   StrPCopy(@FReportFile, '');
   FReportTitle := '';                (* '"No Report Title"' *)
   FMinorTitle := '';
   StrPCopy(@FWindowCaption, '');
   FWindowLocal := wpDesign;
   FWindowStyle := wslNormal;
   FWindowState := wstNormal;
   FWindow := Rect(10,10,300,300);
   FWindowArea := FWindow;
   FLocation := TStringList.Create;
   FLocation.Clear;
   FLocationOutOfDate := True;
   FSelFormula := TStringList.Create;
   FSelFormula.Clear;
   FErrorCode := 0;
   FErrorMessage := '';
 end;

destructor TCrystalVCL.Destroy;
 begin
   (* Crystal Report Engine Specific Data                              *)
   (* I assume PECloseEngine is error tolerant *)
   PECloseEngine;
   FLocation.Clear;
   FLocation.Free;
   FSelFormula.Clear;
   FSelFormula.Free;
   (* Windows Printer Zone Specific Routines                           *)
   ClearPrinterDevice;
   (* Inherited Feature                                                *)
   inherited Destroy;
 end;


(* ************************************************************************** *)
(* ************************************************************************** *)
(* DecodeSTR ---------------------------------------------------------
   Returns a pointer to the first comma delimited field pointed to
   by Str. It replaces the comma with a #0 and moves the Str to the
   beginning of the next string (skipping white space).  Str will
   will point to a #0 character if no more strings are left.  This
   routine is used to fetch strings out of text retrieved from
   WIN.INI.                                                            *)
function TCrystalVCL.DecodeSTR(var Str: PChar): PChar;
 begin
   Result := Str;
   if (Str <> nil) then
     begin
       while (Str^ <> #0) and (Str^ <> ',') do Str := AnsiNext(Str);
       if (Str^ <> #0) then
         begin
           Str^ := #0;
           Inc(Str);
           while Str^ = ' ' do Str := AnsiNext(Str);
         end;
     end;
 end;

(* Clears the association of this object with the current device       *)
procedure TCrystalVCL.ClearPrinterDevice;
 begin
   StrDispose(FDevice);
   FDevice := nil;
   StrDispose(FDriver);
   FDriver := nil;
   StrDispose(FPort);
   FPort := nil;
   if (FDeviceModule >= 32) then
     begin
       FreeLibrary(FDeviceModule);
       FDeviceModule := 0;
     end;
   if (FDevSettings <> nil) and (FDevSettingSize >= 0) then
     FreeMem(FDevSettings, FDevSettingSize);
   FStatus := psUnassociated;
 end;

(* Associates the printer object with a new device. If the ADevice
   parameter is nil the Windows default printer is used, otherwise,
   the parameters must be ones contained in the [devices] section
   of the WIN.INI file.                                                *)
procedure TCrystalVCL.GetPrinterDeviceSpec(ADevice, ADriver, APort: PChar);
 var
   DriverName           : array[0..80] of Char;
   StubDevMode          : TDevMode;
   DevModeSize          : Integer;
  (* ------ Sub Procedure : Get Default Printer ---------------------  *)
  procedure GetDefaultPrinter;
   var
     Printer            : array[0..80] of Char;
     CurrPrinter        : PChar;
   begin
     GetProfileString('windows','device','',Printer,(SizeOf(Printer)-1));
     CurrPrinter := Printer;
     FDevice := StrNew(DecodeSTR(CurrPrinter));
     FDriver := StrNew(DecodeSTR(CurrPrinter));
     FPort := StrNew(DecodeSTR(CurrPrinter));
   end;
  (* ------ Sub Procedure : Equal Null-Term String ------------------  *)
  function EqualPStr(S1, S2: PChar): Boolean;
   begin
     Result := (S1 <> nil) and (S2 <> nil) and (StrComp(S1, S2) = 0);
   end;
 (* -- Main Procedure : SetDevice -----------------------------------  *)
 begin
   if not (EqualPStr(FDevice, ADevice) and EqualPStr(FDriver, ADriver) and
           EqualPStr(FPort, APort)) then
     begin
       ClearPrinterDevice;
       if ADevice = nil then GetDefaultPrinter
        else begin
               FDevice := StrNew(ADevice);
               FDriver := StrNew(ADriver);
               FPort := StrNew(APort);
             end;
       if (FDevice <> nil) and (FDriver <> nil) and (FPort <> nil) then
         begin
           FStatus := psOk;
           StrLCopy(DriverName, FDriver, SizeOf(DriverName) - 1);
           StrLCat(DriverName, '.DRV', SizeOf(DriverName) - 1);
           FDeviceModule := LoadLibrary(DriverName);
           if (FDeviceModule < 32) then FStatus := psInvalidDevice
            else begin         (* Grab the DevMode procedures  *)
                   @FExtDeviceMode := GetProcAddress(FDeviceModule, 'ExtDeviceMode');
                   @FDeviceMode := GetProcAddress(FDeviceModule, 'DeviceMode');
                   if (@FDeviceMode = nil) and (@FExtDeviceMode = nil) then
                     FStatus := psInvalidDevice;
                   if (@FExtDeviceMode <> nil) then
                     begin     (* Get default printer settings *)
                       FDevSettingSize := FExtDeviceMode(0, FDeviceModule, StubDevMode,
                                                        FDevice, FPort, StubDevMode,
                                                        nil, 0);
                       GetMem(FDevSettings, FDevSettingSize);
                       FExtDeviceMode(0, FDeviceModule, FDevSettings^, FDevice, FPort,
                                      FDevSettings^, nil, dm_Out_Buffer);
                     end
                    else FDevSettings := nil;   (* Cannot use local settings *)
                 end;
         end
        else FStatus := psUnassociated;
     end;
 end;

function TCrystalVCL.UpdatePrinterDeviceSpec : Boolean;
 var
   ZDevice, ZDriver, ZPort      : array[0..31] of char;
   ZDeviceMode                  : THandle;
 begin
   try
    Printer.GetPrinter(ZDevice,ZDriver,ZPort,ZDeviceMode);
    if (FAUPrinterSpec) then GetPrinterDeviceSpec(ZDevice,ZDriver,ZPort)
     else GetPrinterDeviceSpec(nil,nil,nil);
   finally
    Result := (FStatus = psOk);
   end;
 end;

(* ************************************************************************** *)
(* ************************************************************************** *)
function TCrystalVCL.GetReportFile : string;
 begin
   Result := StrPas(@FReportFile);
 end;

function TCrystalVCL.GetReportTitle(Index: Integer) : string;
 begin
   case Index of
     0: Result := FReportTitle;
     1: Result := FMinorTitle;
    end;
 end;

function TCrystalVCL.GetWindowTitle : string;
 begin
   Result := StrPas(@FWindowCaption);
 end;

function TCrystalVCL.GetWindowItem(Index: Integer) : Longint;
 begin
   case Index of
     0: Result := FWindow.Top;
     1: Result := FWindow.Left;
     2: Result := FWindow.Right;   (* Read: Width  *)
     3: Result := FWindow.Bottom;  (* Read: Height *)
    end;
 end;

procedure TCrystalVCL.SetReportFile(DReportFile: string);
 begin
   StrPCopy(@FReportFile, DReportFile);
 end;

procedure TCrystalVCL.SetReportTitle(Index: Integer; DTitle: string);
 begin
   case Index of
     0: FReportTitle := DTitle;
     1: FMinorTitle := DTitle;
    end;
 end;

procedure TCrystalVCL.SetWindowTitle(DTitle: string);
 begin
   StrPCopy(@FWindowCaption, DTitle);
 end;

procedure TCrystalVCL.SetWindowLocal(DValue: TXWinLocal);
 begin
   FWindowLocal := DValue;
   case FWindowLocal of
     wpDefault : FWindowArea := Rect(WSE_USEDEFAULT,WSE_USEDEFAULT,WSE_USEDEFAULT,WSE_USEDEFAULT);
     wpDesign  : FWindowArea := FWindow;
    end;
 end;

procedure TCrystalVCL.SetWindowStyle(DValue: TXWinStyle);
 var
  CVOwner            : TComponent;
 begin
   CVOwner := Self.Owner;
   if (CVOwner is TForm) then FWindowStyle := DValue
    else FWindowStyle := wslNormal;
 end;

procedure TCrystalVCL.SetWindowItem(Index: Integer; DValue: Longint);
 begin
   case Index of
     0: FWindow.Top := DValue;
     1: FWindow.Left := DValue;
     2: FWindow.Right := DValue;    (* Read: Width  *)
     3: FWindow.Bottom := DValue;   (* Read: Height *)
    end;
 end;

procedure TCrystalVCL.SetString(Index: Integer; Value: TStrings);
 begin
   case Index of
     0: begin
          FLocation.Assign(Value);
          FLocationOutOfDate := True;
        end;
     1: FSelFormula.Assign(Value);
    end;
 end;


(* ************************************************************************** *)
(* ************************************************************************** *)
function TCrystalVCL.Execute : Boolean;
 var
   PrintJob              : Integer;
   CVWindowHandle        : HWnd;
   CVWindowState         : LongInt;
   IsSuccessful          : Boolean;
   PrinterSelectSuccess  : Boolean;
 begin
   CVWindowHandle := 0;                       (* Determine Destination Window *)
   if (FWindowStyle = wslMDIChild) then
     CVWindowHandle := TForm(Self.Owner).Handle;
   CVWindowState := WSE_BASIS;                      (* Determine Window State *)
   case FWindowState of
     wstNormal    : CVWindowState := (CVWindowState + 0);
     wstMinimized : CVWindowState := (CVWindowState + WSE_MINIMIZE);
     wstMaximized : CVWindowState := (CVWindowState + WSE_MAXIMIZE);
    end;
   PrinterSelectSuccess := UpdatePrinterDeviceSpec;          (* Init. Printer *)
   if (FLocationOutOfDate) then FBDEerror := ConvertAliases; (* Update Locals *)
   FCREerror := False;
   if (FOnLine and (not FBDEerror)) then                     (* Create Report *)
    begin                                                    (* ...           *)
      PrintJob := PEOpenPrintJob(@FReportFile);
      if (PrintJob <> 0) then
       try
        IsSuccessful := PEOutputToWindow(PrintJob, @FWindowCaption,
                                          FWindowArea.Left, FWindowArea.Top, FWindowArea.Right,
                                          FWindowArea.Bottom, CVWindowState, CVWindowHandle);
        if (FReportTitle <> '') then UpdateSmallFormula(PrintJob, 'FCRepTitle', FReportTitle);
        if (FMinorTitle <> '') then UpdateSmallFormula(PrintJob, 'FCRepSecTitle', FMinorTitle);
        if (FSelFormula.Count > 0) then UpdateSelectionFormula(PrintJob);
        if (IsSuccessful and (FLocation.Count > 0)) then
          IsSuccessful := UpdateReportLocation(PrintJob);
        if (IsSuccessful and PrinterSelectSuccess) then
          IsSuccessful := PESelectPrinter(PrintJob, FDriver, FDevice, FPort, FDevSettings^);
        if (IsSuccessful) then
          IsSuccessful := PEStartPrintJob(PrintJob, True);
       finally
        FCREerror := (not IsSuccessful);
        if (FCREerror) then DetermineError(PrintJob);
        PEClosePrintJob(PrintJob);
       end;
    end;
   Result := (FErrorCode = 0);
 end;


(* ************************************************************************** *)
(* ************************************************************************** *)

function TCrystalVCL.ConvertAliases : Boolean;
 var
   SRLoop, SCLoop        : Integer;
   ItemDIV               : Integer;
   ItemSTR               : string;
   OldLocale, NewLocale  : string;
 begin
   Result := True;
   try
    FLocation.BeginUpdate;
    for SRLoop := 0 to (FLocation.Count - 1) do
     begin
       ItemSTR := UpperCase(FLocation.Strings[SRLoop]);
       ItemDIV := POS('|', ItemSTR);
       OldLocale := Copy(ItemSTR, 0, ItemDIV);
       NewLocale := Copy(ItemSTR, (ItemDIV + 1), (Length(ItemSTR) - ItemDIV));
       if ((POS(':',NewLocale) = 0) and (POS('\',NewLocale) = 0)) then
        begin
          NewLocale := GetAliasLocation(NewLocale);
          if (POS(':',NewLocale) > 0) then
            FLocation.Strings[SRLoop] := (OldLocale + NewLocale);
        end;
     end;
    Result := False;
    FLocationOutOfDate := False;
   finally
    FLocation.EndUpdate;
   end;
 end;

function TCrystalVCL.GetAliasLocation(const AliasName: string) : string;
 var
   SDAlias            : array[0..31] of Char;
   AliasDESC          : DBDesc;
 begin
   Result := '';
   try
    StrPLCopy(SDAlias, AliasName, (SizeOf(SDAlias) - 1));
    AnsiToOem(SDAlias, SDAlias);       (* NEXT LINE:  CheckAliasDetail *)
    if (DbiGetDatabaseDesc(SDAlias, @AliasDESC) <> 0) then FBDEerror := True;
    if (StrIComp(AliasDESC.szDbType, 'STANDARD') <> 0) then FBDEerror := True
     else begin
            OemToAnsi(AliasDESC.szPhyName, AliasDESC.szPhyName);
            Result := Format('%s', [AliasDESC.szPhyName]);
          end;
   finally
   end;
 end;

procedure TCrystalVCL.DetermineError(var printJob: Integer);
 begin
   try
    FErrorCode := PEGetErrorCode(printJob);
    case FErrorCode of
        0 : FErrorMessage := '';
      500 : FErrorMessage := 'The System has insufficient memory available';
      501 : FErrorMessage := 'The Job Number specified does not exist';
      502 : FErrorMessage := 'Parameter error: Invalid handle';
      503 : FErrorMessage := 'Return error: String too long';
      504 : FErrorMessage := 'The specified Report File [path] does not exit';
      505 : FErrorMessage := 'No destination was specified';
      507 : FErrorMessage := 'The specified File Name is invalid';
      512 : FErrorMessage := 'The Crystal Report Engine is not open';
      513 : FErrorMessage := 'The specified Printer is invalid';
      522 : FErrorMessage := 'The System has insufficient resources available';
      525 : FErrorMessage := 'The specified Report File has corrupted';
      526 : FErrorMessage := 'No Default Printer is set';
      529 : FErrorMessage := 'The disk is full';
      532 : FErrorMessage := 'One of the required database DLLs has corrupted';
      533 : FErrorMessage := 'A report database has possibly corrupted';
      534 : FErrorMessage := 'A report database has possibly corrupted';
     else
      FErrorMessage := 'An error occured while attempting to print the report';
     end;
   finally;
   end;
 end;


(* ************************************************************************** *)
(* ************************************************************************** *)

function TCrystalVCL.UpdateReportLocation(var printJob: Integer) : Boolean;
 var
   DBLocale              : PETableLocation;
   TBLCount, CRLoop      : Integer;
   SRLoop, SCLoop        : Integer;
   ItemDIV               : Integer;
   Successful            : Boolean;
   FilePath, FileName    : string;
   ItemSTR, FileAddress  : string;
   OldLocale, NewLocale  : string;
 begin
   FCREerror := False;
   if (FLocation.Count > 0) then
    try
     TBLCount := PEGetNTables(printJob);
     if (TBLCount = -1) then FCREerror := True
      else for CRLoop := 0 to (TBLCount - 1) do
            begin
              DBLocale.StructSize := 258;
              Successful := PEGetNthTableLocation(printJob, CRLoop, DBLocale);
              if (Successful) then
               begin
                 FileAddress := StrPas(@DBLocale.Location);
                 GetAddressComponents(FileAddress, FileName, FilePath);
                 FilePath := GetNewLocation(FilePath);
                 FileAddress := (FilePath + '\' + FileName);
                 StrPCopy(@DBLocale.Location, FileAddress);
                 DBLocale.StructSize := 258; { (StrLen(@DBLocale.Location)); }
                 Successful := PESetNthTableLocation(printJob, CRLoop, DBLocale);
               end;
            end;
    finally
    end;
   Result := (FCREerror = False);
 end;

procedure TCrystalVCL.GetAddressComponents(DAddress: string; var DName: string;
                                           var DPath: string);
 var
   BarPOS      : Integer;
 begin
   DName := DAddress;
   BarPOS := POS('\', DName);
   while (BarPOS > 0) do
    begin
      DName := Copy(DName, (BarPOS + 1), (Length(DName) - BarPOS));
      BarPOS := POS('\', DName);
    end;
   DPath := Copy(DAddress, 1, (Length(DAddress) - Length(DName) - 1));
 end;

function TCrystalVCL.GetNewLocation(FilePath: string) : string;
 var
   SRLoop, SCLoop        : Integer;
   ItemDIV               : Integer;
   ItemSTR               : string;
   OldLocale, NewLocale  : string;
 begin
   Result := FilePath;
   for SRLoop := 0 to (FLocation.Count - 1) do
    begin
      ItemSTR := UpperCase(FLocation.Strings[SRLoop]);
      ItemDIV := POS('|', ItemSTR);
      OldLocale := UpperCase(Copy(ItemSTR, 0, (ItemDIV - 1)));
      NewLocale := Copy(ItemSTR, (ItemDIV + 1), (Length(ItemSTR) - ItemDIV));
      if (OldLocale = UpperCase(FilePath)) then
        Result := NewLocale;
    end;
 end;


(* ************************************************************************** *)
(* ************************************************************************** *)

function TCrystalVCL.UpdateSmallFormula(var printJob: Integer;
                                        DFName, DFormula: string) : Boolean;
 var
   SRFName, SRFormula    : string[255];
 begin
   Result := True;
   if (Length(DFormula) > 0) then
    try
     if (DFormula[1] <> '"') then DFormula := ('"' + DFormula);
     if (DFormula[Length(DFormula)] <> '"') then DFormula := (DFormula + '"');
     StrPCopy(@SRFName, DFName);
     StrPCopy(@SRFormula, DFormula);
     PESetFormula(printJob, @SRFName, @SRFormula);
    finally
    end;
 end;


(* ************************************************************************** *)
(* ************************************************************************** *)

function TCrystalVCL.UpdateSelectionFormula(var printJob: Integer) : Boolean;
 var
   SRNLine               : array[0..258] of char;
   PCharItem, MemHOLDER  : PChar;
   SRLoop                : Integer;
   ItemSize              : LongInt;
 begin
   Result := True;
   ItemSize := (GetStringSize(FSelFormula) + 10);
   if (ItemSize <= 65000) then
    try
     MemHOLDER := StrAlloc(ItemSize);
     StrPCopy(MemHOLDER, '');
     PCharItem := MemHOLDER;
     for SRLoop := 0 to (FSelFormula.Count - 1) do
      begin
        StrPCopy(SRNLine, (FSelFormula.Strings[SRLoop] + ' '));
        PCharItem := StrCat(PCharItem, SRNLine);
      end;
     (*   Result := PESetSelectionFormula(printJob, '({pick.EmpID} > 0)');  *)
     if (StrLen(PCharItem) > 1) then Result := PESetSelectionFormula(printJob, PCharItem);
    finally
     StrDispose(MemHOLDER);
     PCharItem := nil;
     if (not PECheckSelectionFormula(printJob)) then PESetSelectionFormula(printJob, '');
    end;
 end;

function TCrystalVCL.GetStringSize(DStringList: TStrings) : LongInt;
 var
   SRLoop                : Integer;
   SRSize                : LongInt;
 begin
   SRSize := 0;
   for SRLoop := 0 to (DStringList.Count - 1) do
     SRSize := (SRSize + Length(DStringList.Strings[SRLoop] + ' '));
   Result := SRSize;
 end;


(* ************************************************************************** *)
(* ************************************************************************** *)
(* ************************************************************************** *)

(* ************************************************************************** *)
(* ************************************************************************** *)
(* ************************************************************************** *)

end.
